@@ -4,6 +4,7 @@
 from __future__ import absolute_import
 import itertools
 import mpd
+import mpd.asyncio
 import os
 import socket
 import sys
@@ -11,6 +12,7 @@ import types
 import warnings
 
 import unittest
+from unittest import mock
 
 try:
     from twisted.python.failure import Failure
@@ -23,16 +25,8 @@ except ImportError:
     )
     TWISTED_MISSING = True
 
-if sys.version_info >= (3,):
-    import asyncio
-else:
-    asyncio = None
+import asyncio
 
-try:
-    import mock
-except ImportError:
-    print("Please install mock from PyPI to run tests!")
-    sys.exit(1)
 
 # show deprecation warnings
 warnings.simplefilter("default")
@@ -57,7 +51,7 @@ class TestMPDClient(unittest.TestCase):
             # Create a new socket.socket() mock with default attributes,
             # each time we are calling it back (otherwise, it keeps set
             # attributes across calls).
-            # That's probablyy what we want, since reconnecting is like
+            # That's probably what we want, since reconnecting is like
             # reinitializing the entire connection, and so, the mock.
             mock.MagicMock(name="socket.socket")
         )
@@ -103,10 +97,6 @@ class TestMPDClient(unittest.TestCase):
         # space should be replaced
         self.assertFalse(hasattr(self.client, "sticker get"))
         self.assertTrue(hasattr(self.client, "sticker_get"))
-        # each command should have prefixe
-        self.assertTrue(hasattr(self.client, "close"))
-        self.assertTrue(hasattr(self.client, "fetch_close"))
-        self.assertTrue(hasattr(self.client, "send_close"))
 
     def test_duplicate_tags(self):
         self.MPDWillReturn("Track: file1\n", "Track: file2\n", "OK\n")
@@ -208,16 +198,6 @@ class TestMPDClient(unittest.TestCase):
         self.assertEqual("0", e["pos"])
         self.assertEqual("66", e["id"])
 
-    def test_send_and_fetch(self):
-        self.MPDWillReturn("volume: 50\n", "OK\n")
-        result = self.client.send_status()
-        self.assertEqual(None, result)
-        self.assertMPDReceived("status\n")
-
-        status = self.client.fetch_status()
-        self.assertEqual(1, self.client._wfile.write.call_count)
-        self.assertEqual({"volume": "50"}, status)
-
     def test_readcomments(self):
         self.MPDWillReturn(
             "major_brand: M4V\n", "minor_version: 1\n", "lyrics: Lalala\n", "OK\n"
@@ -240,63 +220,48 @@ class TestMPDClient(unittest.TestCase):
             self.assertEqual("0", song["pos"])
             self.assertEqual("66", song["id"])
 
-    def test_idle(self):
-        self.MPDWillReturn("OK\n")  # nothing changed after idle-ing
-        self.client.idletimeout = 456
-        res = self.client.idle()
-        self.assertMPDReceived("idle\n")
-        self.client._sock.settimeout.assert_has_calls([mock.call(456), mock.call(None)])
-        self.assertEqual([], res)
-
-        self.client.send_idle()
-        # new event
-        self.MPDWillReturn("changed: update\n", "OK\n")
-
-        event = self.client.fetch_idle()
-        self.assertEqual(event, ["update"])
-
-    def test_noidle(self):
-        self.MPDWillReturn("OK\n")  # nothing changed after idle-ing
-        self.client.send_idle()
-        self.MPDWillReturn("OK\n")  # nothing changed after noidle
-        self.assertEqual(self.client.noidle(), [])
-        self.assertMPDReceived("noidle\n")
-        self.MPDWillReturn("volume: 50\n", "OK\n")
-        self.client.status()
-        self.assertMPDReceived("status\n")
-
-    def test_noidle_while_idle_started_sending(self):
-        self.MPDWillReturn("OK\n")  # nothing changed after idle-ing
-        self.client.send_idle()
-        self.MPDWillReturn("changed: player\n", "OK\n")  # noidle response
-        self.assertEqual(self.client.noidle(), ["player"])
-        self.MPDWillReturn("volume: 50\n", "OK\n")
-        status = self.client.status()
-        self.assertEqual({"volume": "50"}, status)
-
-    def test_throw_when_calling_noidle_withoutidling(self):
-        self.assertRaises(mpd.CommandError, self.client.noidle)
-        self.client.send_status()
-        self.assertRaises(mpd.CommandError, self.client.noidle)
-
     def test_add_and_remove_command(self):
         self.MPDWillReturn("ACK awesome command\n")
 
         self.client.add_command("awesome command", mpd.MPDClient._parse_nothing)
         self.assertTrue(hasattr(self.client, "awesome_command"))
-        self.assertTrue(hasattr(self.client, "send_awesome_command"))
-        self.assertTrue(hasattr(self.client, "fetch_awesome_command"))
         # should be unknown by mpd
         self.assertRaises(mpd.CommandError, self.client.awesome_command)
 
         self.client.remove_command("awesome_command")
         self.assertFalse(hasattr(self.client, "awesome_command"))
-        self.assertFalse(hasattr(self.client, "send_awesome_command"))
-        self.assertFalse(hasattr(self.client, "fetch_awesome_command"))
 
         # remove non existing command
         self.assertRaises(ValueError, self.client.remove_command, "awesome_command")
 
+    def test_partitions(self):
+        self.MPDWillReturn("partition: default\n", "partition: partition2\n", "OK\n")
+        partitions = self.client.listpartitions()
+        self.assertMPDReceived("listpartitions\n")
+        self.assertEqual(
+            [
+                {"partition": "default"},
+                {"partition": "partition2"},
+            ],
+            partitions
+        )
+
+        self.MPDWillReturn("OK\n")
+        self.assertIsNone(self.client.newpartition("Another Partition"))
+        self.assertMPDReceived('newpartition "Another Partition"\n')
+
+        self.MPDWillReturn("OK\n")
+        self.assertIsNone(self.client.partition("Another Partition"))
+        self.assertMPDReceived('partition "Another Partition"\n')
+
+        self.MPDWillReturn("OK\n")
+        self.assertIsNone(self.client.delpartition("Another Partition"))
+        self.assertMPDReceived('delpartition "Another Partition"\n')
+
+        self.MPDWillReturn("OK\n")
+        self.assertIsNone(self.client.moveoutput("My ALSA Device"))
+        self.assertMPDReceived('moveoutput "My ALSA Device"\n')
+
     def test_client_to_client(self):
         # client to client is at this time in beta!
 
@@ -328,21 +293,10 @@ class TestMPDClient(unittest.TestCase):
         self.assertEqual([], channels)
 
     def test_unicode_as_command_args(self):
-        if sys.version_info < (3, 0):
-            self.MPDWillReturn("OK\n")
-            res = self.client.find("file", unicode("☯☾☝♖✽", "utf-8"))
-            self.assertIsInstance(res, list)
-            self.assertMPDReceived('find "file" "☯☾☝♖✽"\n')
-
-            self.MPDWillReturn("OK\n")
-            res2 = self.client.find("file", "☯☾☝♖✽")
-            self.assertIsInstance(res2, list)
-            self.assertMPDReceived('find "file" "☯☾☝♖✽"\n')
-        else:
-            self.MPDWillReturn("OK\n")
-            res = self.client.find("file", "☯☾☝♖✽")
-            self.assertIsInstance(res, list)
-            self.assertMPDReceived('find "file" "☯☾☝♖✽"\n')
+        self.MPDWillReturn("OK\n")
+        res = self.client.find("file", "☯☾☝♖✽")
+        self.assertIsInstance(res, list)
+        self.assertMPDReceived('find "file" "☯☾☝♖✽"\n')
 
     def test_numbers_as_command_args(self):
         self.MPDWillReturn("OK\n")
@@ -754,7 +708,7 @@ class TestMPDClientSocket(unittest.TestCase):
 
         # we're expecting a timeout
         self.assertRaises(
-            mpd.ConnectionError, lambda: self.client.albumart("a/full/path.mp3")
+            socket.timeout, lambda: self.client.albumart("a/full/path.mp3")
         )
 
         self.assertMPDReceived(
@@ -767,7 +721,7 @@ class TestMPDClientSocket(unittest.TestCase):
 
         # we're expecting a timeout or error of some form
         self.assertRaises(
-            mpd.ConnectionError, lambda: self.client.albumart("a/full/path.mp3")
+            socket.timeout, lambda: self.client.albumart("a/full/path.mp3")
         )
 
         self.assertMPDReceived(b'albumart "a/full/path.mp3" "0"\n')
@@ -788,7 +742,7 @@ class TestMPDClientSocket(unittest.TestCase):
 
         real_binary = self.client.albumart("a/full/path.mp3")
         self.assertMPDReceived(b'albumart "a/full/path.mp3" "0"\n')
-        self.assertEqual(real_binary, expected_binary)
+        self.assertEqual(real_binary, {"binary": expected_binary})
 
     def test_binary_albumart_singlechunk_nosize(self):
         # length: 16
@@ -800,7 +754,7 @@ class TestMPDClientSocket(unittest.TestCase):
 
         real_binary = self.client.albumart("a/full/path.mp3")
         self.assertMPDReceived(b'albumart "a/full/path.mp3" "0"\n')
-        self.assertEqual(real_binary, expected_binary)
+        self.assertEqual(real_binary, {"binary": expected_binary})
 
     def test_binary_albumart_singlechunk_sizeheader(self):
         # length: 16
@@ -814,7 +768,7 @@ class TestMPDClientSocket(unittest.TestCase):
 
         real_binary = self.client.albumart("a/full/path.mp3")
         self.assertMPDReceived(b'albumart "a/full/path.mp3" "0"\n')
-        self.assertEqual(real_binary, expected_binary)
+        self.assertEqual(real_binary, {"binary": expected_binary})
 
     def test_binary_albumart_even_multichunk(self):
         # length: 16 each
@@ -846,7 +800,7 @@ class TestMPDClientSocket(unittest.TestCase):
             b'albumart "a/full/path.mp3" "0"\nalbumart "a/full/path.mp3" "16"'
             b'\nalbumart "a/full/path.mp3" "32"\n'
         )
-        self.assertEqual(real_binary, expected_binary)
+        self.assertEqual(real_binary, {"binary": expected_binary})
 
     def test_binary_albumart_odd_multichunk(self):
         # lengths: 17, 15, 1
@@ -875,7 +829,7 @@ class TestMPDClientSocket(unittest.TestCase):
             b'albumart "a/full/path.mp3" "0"\nalbumart "a/full/path.mp3" "17"\n'
             b'albumart "a/full/path.mp3" "32"\n'
         )
-        self.assertEqual(real_binary, expected_binary)
+        self.assertEqual(real_binary, {"binary": expected_binary})
 
     # MPD server can return empty response if a file exists but is empty
     def test_binary_albumart_emptyresponse(self):
@@ -883,7 +837,91 @@ class TestMPDClientSocket(unittest.TestCase):
 
         real_binary = self.client.albumart("a/full/path.mp3")
         self.assertMPDReceived(b'albumart "a/full/path.mp3" "0"\n')
-        self.assertEqual(real_binary, b"")
+        self.assertEqual(real_binary, {"binary": b""})
+
+    # readpicture returns empty object if the song exists but has no picture
+    def test_binary_readpicture_emptyresponse(self):
+        self.MPDWillReturnBinary(b"OK\n")
+
+        real_binary = self.client.readpicture("plainsong.mp3")
+        self.assertMPDReceived(b'readpicture "plainsong.mp3" "0"\n')
+        self.assertEqual(real_binary, {})
+
+    def test_binary_readpicture_untyped(self):
+        # length: 16 each
+        expected_chunk1 = (
+            b"\x01\x02\x00\x03\x04\x00\xFF\x05\x07\x08\x0A\x0F\xF0\xA5\x00\x01"
+        )
+        expected_chunk2 = (
+            b"\x0A\x0B\x0C\x0D\x0E\x0F\x10\x1F\x2F\x2D\x33\x0D\x00\x00\x11\x13"
+        )
+        expected_chunk3 = (
+            b"\x99\x88\x77\xDD\xD0\xF0\x20\x70\x71\x17\x13\x31\xFF\xFF\xDD\xFF"
+        )
+        expected_binary = expected_chunk1 + expected_chunk2 + expected_chunk3
+
+        # 3 distinct commands expected
+        self.MPDWillReturnBinary(
+            b"size: 48\nbinary: 16\n"
+            + expected_chunk1
+            + b"\nOK\nsize: 48\nbinary: 16\n"
+            + expected_chunk2
+            + b"\nOK\nsize: 48\nbinary: 16\n"
+            + expected_chunk3
+            + b"\nOK\n"
+        )
+
+        real_binary = self.client.readpicture("a/full/path.mp3")
+
+        self.assertMPDReceived(
+            b'readpicture "a/full/path.mp3" "0"\nreadpicture "a/full/path.mp3" "16"'
+            b'\nreadpicture "a/full/path.mp3" "32"\n'
+        )
+        self.assertEqual(real_binary, {"binary": expected_binary})
+
+    def test_binary_readpicture_typed(self):
+        # length: 16 each
+        expected_binary = bytes(range(48))
+
+        # 3 distinct commands expected
+        self.MPDWillReturnBinary(
+            b"size: 48\ntype: image/png\nbinary: 16\n"
+            + expected_binary[0:16]
+            + b"\nOK\nsize: 48\ntype: image/png\nbinary: 16\n"
+            + expected_binary[16:32]
+            + b"\nOK\nsize: 48\ntype: image/png\nbinary: 16\n"
+            + expected_binary[32:48]
+            + b"\nOK\n"
+        )
+
+        real_binary = self.client.readpicture("a/full/path.mp3")
+
+        self.assertMPDReceived(
+            b'readpicture "a/full/path.mp3" "0"\nreadpicture "a/full/path.mp3" "16"'
+            b'\nreadpicture "a/full/path.mp3" "32"\n'
+        )
+        self.assertEqual(real_binary, {"binary": expected_binary, "type": "image/png"})
+
+    def test_binary_readpicture_badheaders(self):
+        expected_binary = bytes(range(32))
+
+        # inconsistent type header from response 1 to response 2
+        # exception is expected
+        self.MPDWillReturnBinary(
+            b"size: 32\ntype: image/jpeg\nbinary: 16\n"
+            + expected_binary[0:16]
+            + b"\nOK\nsize: 32\ntype: image/png\nbinary: 16\n"
+            + expected_binary[16:32]
+            + b"\nOK\n"
+        )
+
+        self.assertRaises(
+            mpd.CommandError, lambda: self.client.readpicture("song.mp3")
+        )
+
+        self.assertMPDReceived(
+            b'readpicture "song.mp3" "0"\nreadpicture "song.mp3" "16"\n'
+        )
 
 
 class MockTransport(object):
@@ -1172,6 +1210,12 @@ class AsyncMockServer:
         # directly passing around the awaitable
         return self._output.get()
 
+    async def readexactly(self, length):
+        ret = await self._output.get()
+        if len(ret) != length:
+            self.error("Mock data is not chuncked in the way the client expects to read it")
+        return ret
+
     def write(self, data):
         try:
             next_write = self._expectations[0][0][0]
@@ -1201,11 +1245,8 @@ class AsyncMockServer:
         self._feed()
 
 
-@unittest.skipIf(asyncio is None, "requires asyncio to be available")
 class TestAsyncioMPD(unittest.TestCase):
     def init_client(self, odd_hello=None):
-        import mpd.asyncio
-
         self.loop = asyncio.get_event_loop()
 
         self.mockserver = AsyncMockServer()
@@ -1221,12 +1262,19 @@ class TestAsyncioMPD(unittest.TestCase):
         self.mockserver.expect_exchange([], hello_lines)
 
         self.client = mpd.asyncio.MPDClient()
-        self._await(self.client.connect(TEST_MPD_HOST, TEST_MPD_PORT, loop=self.loop))
+        self._await(self.client.connect(TEST_MPD_HOST, TEST_MPD_PORT))
 
         asyncio.open_connection.assert_called_with(
-            TEST_MPD_HOST, TEST_MPD_PORT, loop=self.loop
+            TEST_MPD_HOST, TEST_MPD_PORT
         )
 
+    def __del__(self):
+        # Clean up after init_client. (This works for now; if it causes
+        # trouble, change all init_client to a `with init_client` or use
+        # fixtures).
+        if hasattr(self, "client"):
+            self.client.disconnect()
+
     def _await(self, future):
         return self.loop.run_until_complete(future)
 
@@ -1353,6 +1401,90 @@ class TestAsyncioMPD(unittest.TestCase):
         self.init_client()
         self._await(self._test_list())
 
+    async def _test_albumart(self):
+        self.mockserver.expect_exchange(
+            [b'albumart "x.mp3" "0"\n'],
+            [
+                b"size: 32\n",
+                b"binary: 16\n",
+                bytes(range(16)),
+                b"\n",
+                b"OK\n",
+            ]
+        )
+        self.mockserver.expect_exchange(
+            [b'albumart "x.mp3" "16"\n'],
+            [
+                b"size: 32\n",
+                b"binary: 16\n",
+                bytes(range(16)),
+                b"\n",
+                b"OK\n",
+            ],
+        )
+
+        albumart = await self.client.albumart("x.mp3")
+
+        expected = {"binary": bytes(range(16)) + bytes(range(16))}
+
+        self.assertEqual(albumart, expected)
+
+    async def _test_readpicture(self):
+            self.mockserver.expect_exchange(
+                [b'readpicture "x.mp3" "0"\n'],
+                [
+                    b"size: 32\n",
+                    b"type: image/jpeg\n",
+                    b"binary: 16\n",
+                    bytes(range(16)),
+                    b"\n",
+                    b"OK\n",
+                ]
+            )
+            self.mockserver.expect_exchange(
+                [b'readpicture "x.mp3" "16"\n'],
+                [
+                    b"size: 32\n",
+                    b"type: image/jpeg\n",
+                    b"binary: 16\n",
+                    bytes(range(16)),
+                    b"\n",
+                    b"OK\n",
+                ],
+            )
+
+            art = await self.client.readpicture("x.mp3")
+
+            expected = {"binary": bytes(range(16)) + bytes(range(16)), "type": "image/jpeg"}
+
+            self.assertEqual(art, expected)
+
+    async def _test_readpicture_empty(self):
+            self.mockserver.expect_exchange(
+                [b'readpicture "x.mp3" "0"\n'],
+                [
+                    b"OK\n",
+                ]
+            )
+
+            art = await self.client.readpicture("x.mp3")
+
+            expected = {}
+
+            self.assertEqual(art, expected)
+
+    def test_albumart(self):
+        self.init_client()
+        self._await(self._test_albumart())
+
+    def test_readpicture(self):
+        self.init_client()
+        self._await(self._test_readpicture())
+
+    def test_readpicture_empty(self):
+        self.init_client()
+        self._await(self._test_readpicture_empty())
+
     def test_mocker(self):
         """Does the mock server refuse unexpected writes?"""
         self.init_client()
@@ -1360,6 +1492,65 @@ class TestAsyncioMPD(unittest.TestCase):
         self.mockserver.expect_exchange([b"expecting odd things\n"], [b""])
         self.assertRaises(AssertionError, self._await, self.client.status())
 
+    async def _test_idle(self):
+        self.mockserver.expect_exchange(
+                [b'idle "database"\n',],
+                [b"ACK [4@0] don't let you yet but you shouldn't care\n"],
+                )
+        self.mockserver.expect_exchange(
+                # code could be smarter and set __in_idle = False and not set
+                # noidle, but it's not *wrong* either
+                [b'noidle\n', b'password "1234"\n',],
+                [b"OK\n"],
+                )
+        self.mockserver.expect_exchange(
+                [b'idle "playlist"\n'],
+                [b'idle: playlist\n', b"OK\n"],
+                )
+        self.mockserver.expect_exchange(
+                [b'noidle\n', b'currentsong\n'],
+                # it's a bit brief but it'll be accepted
+                [b"OK\n", ],
+                )
+        self.mockserver.expect_exchange(
+                [b'idle "playlist"\n'],
+                [b"ACK [4@0] I don't let you observe for no other reason\n"],
+                )
+        self.mockserver.expect_exchange(
+                [b'noidle\n', b'status\n'],
+                [b"ACK [4@0] whatever made the idle failed now fails too\n", ],
+                )
+
+        # clearly longer than IMMEDITE_COMMAND_TIMEOUT
+        await asyncio.sleep(0.5)
+
+        await self.client.password("1234")
+
+        async for changeset in self.client.idle(["playlist"]):
+            # the one regular event sent
+            self.assertEqual(changeset, ["playlist"])
+            break
+
+        await self.client.currentsong()
+
+        async for changeset in self.client.idle(["playlist"]):
+            # watching for regular events but the error also gets shown as an event
+            self.assertEqual(changeset, ["playlist"])
+            break
+
+        try:
+            await self.client.status()
+        except mpd.CommandError:
+            pass
+        else:
+            raise AssertionError("The status should have flushed out the error that made the idle fail in the first place")
+
+        self.client.disconnect()
+
+    def test_idle(self):
+        self.init_client()
+        self._await(self._test_idle())
+
 
 if __name__ == "__main__":
     unittest.main()
